#!/usr/bin/env python
# -*- coding: utf-8 -*-

# A module to ease working with fonts and effects.
# Copyright Gummbum, 4/18/2017. Free software, no warranty express or implied.

import os

import pygame
from pygame import SRCALPHA, Color


__all__ = ['load_font', 'size', 'graduate', 'make_colors', 'render']


def load_font(name, size, bold=False, italic=False):
    """load a Font or SysFont
    
    :param name: file name or sysfont family name
    :param size: point size
    :param bold: boolean (software-rendered bold)
    :param italic: boolean (software-rendered italic)
    :return: pygame.Font
    """
    if os.access(name, os.F_OK):
        font = pygame.font.Font(name, size)
    else:
        font = pygame.font.SysFont(name, size)
    font.set_bold(bold)
    font.set_italic(italic)
    return font


def size(text, font, scolor=None, ocolor=None):
    """dimensions of the text if it were rendered by font, given scolor and/or ocolor
    
    :param text: string
    :param font: pygame.Font
    :param scolor: pygame.Color or None
    :param ocolor: pygame.Color or None
    :return: (width, height)
    """
    sz = font.size(text)
    if scolor:
        sz[1] += 2
    elif ocolor:
        sz[0] += 4
        sz[1] += 4
    return sz


def graduate(img, primary, secondary):
    """modify an image (e.g. a plain rendered font) with gradient
    
    :param img: pygame.Surface
    :param primary: pygame.Color to replace; gradient transitions from this color
    :param secondary: pygame.Color to blend; gradient transitions to this color
    :return: same img, modified in place
    """
    primary, secondary = make_colors(primary, secondary)
    pa = pygame.PixelArray(img)
    # Legacy: PixelArray.shape is not available in pygame 1.9.1
    # cols, rows = pa.shape
    cols, rows = img.get_size()
    step = 1.0 / rows
    pf = 1.0
    sf = 0.0
    pri3 = primary[:3]
    for y in range(rows):
        grad = [int(p * pf + s * sf) for p, s in zip(primary[:3], secondary[:3])]
        grad = Color(*grad)
        row = pa[:, y:y+1]
        # row.replace() does not preserve alpha...
        # row.replace(primary, grad)
        for x, ci in enumerate(row):
            try:
                # pygame 1.9.1 (and 2?)
                c = img.unmap_rgb(ci)
            except TypeError:
                # pygame 1.9.3
                c = img.unmap_rgb(ci[0])
            if c[:3] == pri3:
                grad[3] = c[3]
                row[x] = grad
        pf -= step
        sf += step
    img.unlock()
    return img


def make_colors(*values):
    """convert arbitrary color specs to pygame.Color
    
    Results will contain None if None is passed in.
    
    :param values: varargs; arbitrary color specs (e.g. "red", (255, 0, 0), [255, 0, 0], None, Color("red"))
    :return: [pygame.Color, ...]
    """
    colors = []
    for v in values:
        if v is None:
            colors.append(None)
        elif isinstance(v, Color):
            colors.append(v)
        elif isinstance(v, str):
            colors.append(Color(v))
        elif isinstance(v, (tuple, list)):
            colors.append(Color(*v))
        else:
            colors.append(Color(v))
    return colors


def render(text, font, fgcolor, bgcolor=None, gcolor=None, scolor=None, ocolor=None, colorkey=None, alpha=None, antialias=True):
    """render text, applying various effects
    
    For example color specs see the docstring for make_colors().
    
    :param text: string
    :param font: pygame.Font
    :param fgcolor: color spec; font foreground
    :param bgcolor: color spec; font background
    :param gcolor: color spec; gradient color
    :param scolor: color spec; drop shadow color
    :param ocolor: color spec; outline color
    :param colorkey: color spec; colorkey (see pygame.surface.set_colorkey())
    :param alpha: 0 .. 255 (see pygame.surface.set_alpha())
    :param antialias: boolean; render font antialiased
    :return: pygame.Surface
    """
    fgcolor, bgcolor, gcolor, scolor, ocolor, colorkey = make_colors(fgcolor, bgcolor, gcolor, scolor, ocolor, colorkey)
    if scolor:
        # render with drop shadow
        surf = _render_shadow(text, font, fgcolor, bgcolor, gcolor, scolor, colorkey, antialias)
    elif ocolor:
        # render with outline
        surf = _render_outline(text, font, fgcolor, bgcolor, gcolor, ocolor, colorkey, antialias)
    elif bgcolor is None:
        # render with transparent background
        surf = font.render(text, antialias, fgcolor)
        if gcolor:
            graduate(surf, fgcolor, gcolor)
    else:
        # render with colored background
        surf = font.render(text, antialias, fgcolor, bgcolor)
        if gcolor:
            graduate(surf, fgcolor, gcolor)
        surf.set_colorkey(colorkey)
        surf.set_alpha(alpha)

    return surf


def _render_shadow(text, font, fgcolor, bgcolor, gcolor, scolor, colorkey, antialias):
    """render text with a drop shadow"""
    swidth = 2

    # Set up the target surface
    w, h = font.size(text)
    if bgcolor:
        surf = pygame.Surface((w + swidth, h + swidth))
        surf.fill(bgcolor)
        surf.set_colorkey(colorkey)
    else:
        surf = pygame.Surface((w + swidth, h + swidth), SRCALPHA)

    # Blit the outline
    img = font.render(text, antialias, scolor)
    for x, y in ((swidth // 2, swidth), (swidth, swidth), (swidth, swidth // 2)):
        surf.blit(img, (x, y))

    # Blit the text
    img = font.render(text, antialias, fgcolor)
    if gcolor:
        graduate(img, fgcolor, gcolor)
    surf.blit(img, (0, 0))

    return surf

def _render_outline(text, font, fgcolor, bgcolor, gcolor, ocolor, colorkey, antialias):
    """render text with an outline"""
    owidth = 2

    # Set up the target surface
    w, h = font.size(text)
    if bgcolor:
        surf = pygame.Surface((w + 2 * owidth, h + 2 * owidth))
        surf.fill(bgcolor)
        surf.set_colorkey(colorkey)
    else:
        surf = pygame.Surface((w + 2 * owidth, h + 2 * owidth), SRCALPHA)

    # Blit the outline
    img = font.render(text, antialias, ocolor)
    for x in range(0, owidth + 1):
        for y in range(0, owidth + 1):
            surf.blit(img, (x, y))

    # Blit the text
    img = font.render(text, antialias, fgcolor)
    if gcolor:
        graduate(img, fgcolor, gcolor)
    surf.blit(img, (owidth // 2, owidth // 2))

    return surf


if __name__ == '__main__':
    import random
    import re
    from pygame import KEYDOWN, QUIT, K_SPACE, K_RETURN, K_ESCAPE
    pygame.init()
    screen = pygame.display.set_mode((640, 480))
    screen_rect = screen.get_rect()
    clock = pygame.time.Clock()
    max_fps = 30

    # vary screen bg color
    the_colors = [c for c in pygame.color.THECOLORS.keys() if not re.match(r'^gr[ae]y\d\d', c)]
    auto_clear_secs = 8
    clear_name = 'grey33'
    clear_color = Color(clear_name)
    pygame.display.set_caption('Screen color: {}'.format(clear_name))

    fonts = dict(
        vera=('VeraBd.ttf', 24, 0, 0),
        sans=('sans', 32, 0, 0),
        times=('times', 32, 0, 0))
    texts = [
        ('Shadow: Zippity doo dah', 'vera', 'darkblue', 'lightblue', 'black', None),
        ('Shadow: The little brown fox', 'sans', 'red', 'yellow', 'black', None),
        ('Shadow: Once upon a time...', 'times', 'green', 'yellow', 'black', None),
        ('Shadow: ...there was an angry text.', 'times', 'red', 'grey30', 'black', None),
        ('Outline: Zippity doo dah', 'vera', 'darkblue', 'lightblue', None, 'black'),
        ('Outline: The little brown fox', 'sans', 'yellow', 'red', None, 'black'),
        ('Outline: Once upon a time...', 'times', 'yellow', 'green', None, 'black'),
        ('Outline: ...there was a happy text.', 'times', 'deeppink', 'mistyrose', 'black', None),
    ]
    sprites = []
    y = 50
    for text, font_name, fgcolor, gcolor, scolor, ocolor in texts:
        font = load_font(*fonts[font_name])
        surf = render(text, font, fgcolor, gcolor=gcolor, scolor=scolor, ocolor=ocolor)
        rect = surf.get_rect(centerx=screen_rect.centerx, y=y)
        sprites.append((surf, rect))
        y += rect.h + 10
    ticks = 0
    while 1:
        clock.tick(max_fps)
        ticks += 1
        for e in pygame.event.get():
            if e.type == KEYDOWN:
                if e.key == K_SPACE:
                    ticks = max_fps * auto_clear_secs
                elif e.key == K_RETURN:
                    print('Screen color: {}={}'.format(clear_name, clear_color))
                elif e.key == K_ESCAPE:
                    quit()
            elif e.type == QUIT:
                quit()
        if ticks % (max_fps * auto_clear_secs) == 0:
            clear_name = random.choice(the_colors)
            clear_color = Color(clear_name)
            pygame.display.set_caption('Screen color: {}'.format(clear_name))
        screen.fill(clear_color)
        for s, r in sprites:
            screen.blit(s, r)
        pygame.display.flip()
