#!/usr/bin/env python
# -*- coding: utf-8 -*-

# A demo to interactively showcase all the features.
# Copyright Gummbum, 4/18/2017. Free software, no warranty express or implied.

# You don't have to subclass HUD, but I did to make the code cleaner, and benefit from IDE inspection. It really helps
# in this demo because this is a very busy, complex example.

import pygame
from pygame.locals import *

from .hudlight import HUD


class MyHUD(HUD):
    """
    This class does:
    - adds the HUD items
    - provides methods to update each HUD item (calling code looks neater)
    - a monolithic update_all() method
    """

    def __init__(self, font=None, size=None, bold=None, italic=None, fgcolor=None, bgcolor=None, gcolor=None,
                 ocolor=None, scolor=None):
        super(MyHUD, self).__init__(font, size, bold, italic, fgcolor, bgcolor, ocolor, scolor)
        self.add('font', '1=Font ({}), 2=Size ({})', font, size)
        self.add('style', '3=Bold ({}), 4=Italic ({})', bold, italic)
        self.add('antialias', '5=AntiAlias ({})', self.ANTIALIAS)
        self.add('color', '6=FgColor ({}) G=Gradient ({})', fgcolor, gcolor)
        self.add('colorkey', '7=BgColor ({}), 8=ColorKey ({})', bgcolor, self.COLORKEY)
        self.add('shadow', '9=Shadow ({})', self.SHADOWCOLOR)
        self.add('outline', '0=Outline ({})', self.OUTLINECOLOR)
        self.add('fps', 'FPS {:0.0f}', 0)
        self.add('mouse', 'Mouse {0[0]}, {0[1]}', (0, 0))
        self.add('angle', 'Angle {:0.0f}', 0.0)
        self.add('perf', 'Rendered/sec {}', 0)

    def update_all(self, font, size, bold, italic, antialias, fgcolor, bgcolor, gcolor, colorkey, shadow_color,
                   outline_color, fps, mouse_pos, angle):
        self.update_font(font, size)
        self.update_style(bold, italic)
        self.update_antialias(antialias)
        self.update_color(fgcolor, gcolor)
        self.update_colorkey(bgcolor, colorkey)
        self.update_shadow(shadow_color)
        self.update_outline(outline_color)
        self.update_fps(fps)
        self.update_mouse(mouse_pos)
        self.update_angle(angle)

    def update_font(self, font, size):
        self.update_item('font', font, size)

    def update_style(self, bold, italic):
        self.update_item('style', bold, italic)

    def update_antialias(self, antialias):
        self.update_item('antialias', antialias)

    def update_color(self, fgcolor, gcolor):
        self.update_item('color', fgcolor, gcolor)

    def update_colorkey(self, bgcolor, colorkey):
        self.update_item('colorkey', bgcolor, colorkey)

    def update_shadow(self, scolor):
        self.update_item('shadow', scolor)

    def update_outline(self, ocolor):
        self.update_item('outline', ocolor)

    def update_fps(self, fps):
        self.update_item('fps', fps)

    def update_mouse(self, mouse_pos):
        self.update_item('mouse', mouse_pos)

    def update_angle(self, angle):
        self.update_item('angle', angle)

    def update_perf(self):
        self.update_item('perf', int(self.fps()))


class MyUI(object):
    """
    This class:
    - contains runtime settings that the user can change via key presses
    - creates a keymap: pygame keys -> cycle_*() methods (calling code is much neater; see do_key())
    - circular linked lists for the menus to cycle through (see the circular linked lists at the end of
      __init__ if you wish to customize the menu data; this is pretty easy, but read the notes)
    - cycle_*() methods (do the work of cycling a menu choice via its circular linked list)
    """

    def __init__(self, font, size, bold, italic, antialias, fgcolor, bgcolor, gcolor, colorkey, shadow, outline,
                 keys=None):
        # Current values.
        self.font = font
        self.size = size
        self.bold = bold
        self.italic = italic
        self.antialias = antialias
        self.fgcolor = fgcolor
        self.bgcolor = bgcolor
        self.gcolor = gcolor
        self.colorkey = colorkey
        self.shadow = shadow
        self.outline = outline

        # Keymap: pygame key -> cycle_*() method. do_key() handles the lookups and calls.
        self.keymap = {}
        if keys:
            self.set_keymap(keys)
        self.is_interactive = False

        # Circular linked lists, used by the cycle*() methods.
        # Usage: new_font = next_font.get(current_font, 'default').
        # Customization: see make_cll() docstring for explanation of args.
        # NOTE: make sure there is a default in position 0, even if it repeats position 1. All the code
        # relies on this.
        self.cll_font = self._make_cll(*'VeraBD.ttf VeraBD.ttf sans times'.split(), printme=False)
        self.cll_size = self._make_cll(28, 28, 22, 18, 16, printme=False)
        self.cll_bold = self._make_cll(False, True, False, printme=False)
        self.cll_italic = self._make_cll(False, True, False, printme=False)
        self.cll_antialias = self._make_cll(True, True, False, printme=False)
        self.cll_fgcolor = self._make_cll(*'red red darkorange1 deeppink darkgreen darkblue grey30'.split(),
                                          printme=False)
        self.cll_bgcolor = self._make_cll(None, None, 'midnightblue', 'skyblue', printme=False)
        self.cll_gcolor = self._make_cll(None, None, 'yellow', 'mistyrose', 'white', 'lightblue', printme=False)
        self.cll_colorkey = self._make_cll(None, None, 'midnightblue', 'skyblue', printme=False)
        self.cll_shadow = self._make_cll('black', None, 'black', 'grey', 'white', printme=False)
        self.cll_outline = self._make_cll(None, None, 'black', 'grey', 'white', printme=False)

    def _make_cll(self, *words, **kwargs):
        """Make a circular linked list from arbitrary words

        :param words: contents are: default, choice1, choice2, ...
        :param kwargs: printme=True for a debug print
        :return: None
        """
        printme = kwargs.get('printme', False)
        cll = {}
        null = '__null__'
        init = first = key = null
        for word in words:
            if init is null:
                cll['default'] = word
                init = word
            elif first is null:
                key = first = word
            else:
                cll[key] = word
                key = word
        cll[key] = first
        if printme:
            print([(k, cll[k]) for k in (['default'] + list(words))])
        return cll

    def set_keymap(self, keys):
        # Map the event keys to cycle*() methods.
        names = 'font size bold italic antialias fgcolor bgcolor gcolor colorkey shadow outline'.split()
        methods = [getattr(self, 'cycle_' + n) for n in names]
        self.keymap = dict(zip(keys, methods))

    def do_key(self, key, hud):
        # Look up the key and call its assigned cycle*() method
        if key in self.keymap:
            self.keymap[key](hud)

    def set_interactive(self, hud):
        # Call once to go from initial splash to interactive mode
        if not self.is_interactive:
            self._reset_defaults()
            self._set_font(hud)
            hud.set_fgcolor(self.fgcolor)
            hud.set_bgcolor(self.bgcolor)
            hud.set_gradient(self.gcolor)
            hud.set_colorkey(self.colorkey)
            hud.set_antialias(self.antialias)
            hud.set_shadow(self.shadow)
            hud.set_outline(self.outline)
            self.is_interactive = True

    def cycle_font(self, hud=None):
        self.font = self._cycle(self.cll_font, self.font)
        hud and hud.set_font_name(self.font)
        return self.font

    def cycle_size(self, hud=None):
        self.size = self._cycle(self.cll_size, self.size)
        hud and hud.set_size(self.size)
        return self.size

    def cycle_bold(self, hud=None):
        self.bold = self._cycle(self.cll_bold, self.bold)
        hud and hud.set_bold(self.bold)
        return self.bold

    def cycle_italic(self, hud=None):
        self.italic = self._cycle(self.cll_italic, self.italic)
        hud and hud.set_italic(self.italic)
        return self.italic

    def cycle_antialias(self, hud):
        self.antialias = self._cycle(self.cll_antialias, self.antialias)
        hud.set_antialias(self.antialias)
        return self.antialias

    def cycle_fgcolor(self, hud):
        self.fgcolor = self._cycle(self.cll_fgcolor, self.fgcolor)
        hud.set_fgcolor(self.fgcolor)
        return self.fgcolor

    def cycle_bgcolor(self, hud):
        self.bgcolor = self._cycle(self.cll_bgcolor, self.bgcolor)
        hud.set_bgcolor(self.bgcolor)
        return self.bgcolor

    def cycle_gcolor(self, hud):
        self.gcolor = self._cycle(self.cll_gcolor, self.gcolor)
        hud.set_gradient(self.gcolor)
        return self.gcolor

    def cycle_colorkey(self, hud):
        self.colorkey = self._cycle(self.cll_colorkey, self.colorkey)
        hud.set_colorkey(self.colorkey)
        return self.colorkey

    def cycle_shadow(self, hud):
        self.shadow = self._cycle(self.cll_shadow, self.shadow)
        hud.set_shadow(self.shadow)
        return self.shadow

    def cycle_outline(self, hud):
        self.outline = self._cycle(self.cll_outline, self.outline)
        hud.set_outline(self.outline)
        return self.outline

    def _reset_defaults(self):
        # Set everything to ccl['default'] value
        for name in 'font size bold italic antialias fgcolor bgcolor gcolor colorkey shadow outline'.split():
            cll = getattr(self, 'cll_' + name)
            setattr(self, name, cll['default'])

    def _cycle(self, cll, value):
        return cll.get(value, cll['default'])

    def _set_font(self, hud):
        hud.set_font(self.font, self.size, self.bold, self.italic)


def run_demo():
    import math

    # The basic pygame things
    pygame.init()
    clock = pygame.time.Clock()
    screen = pygame.display.set_mode((900, 600))
    screen_rect = screen.get_rect()
    pygame.display.set_caption('HUD Demo - Welcome. Press a key to start | Q or ESC to quit')
    clear_color = Color('grey20')

    # Some runtime things for the HUD to watch
    origin = screen_rect.center
    radius = screen_rect.centerx * 4 / 5
    angle = 0
    mouse_pos = 0, 0
    max_fps = 60

    # Some decorative lines to break up the monotone background
    def make_lines(lines):
        radius = max(screen_rect.w // 2, screen_rect.h // 2)
        for angle in range(0, 360, 2):
            radians_ = math.radians(angle)
            x = origin[0] + radius * math.cos(radians_)
            y = origin[1] + radius * math.sin(radians_)
            lines.append((origin, (x, y)))

    decorative_lines = []
    make_lines(decorative_lines)

    # Create the user interface. This is driven by key presses, and configures the HUD on the fly.
    # Keymaps: ui.do_key() -> ui.cycle_*()
    ui = MyUI(
        'fixed', 26, True, True, True,  # font, size, bold, italic, antialias
        'hotpink4', None, None,  # fgcolor, bgcolor, gcolor
        None, 'hotpink', None,  # colorkey, shadow, outline
        # keys, see MyUI.set_keymap()
        (K_1, K_2, K_3, K_4, K_5,  # font, size, bold, italic, antialias
         K_6, K_7, K_g,  # fgcolor, bgcolor, gcolor
         K_8, K_9, K_0))  # colorkey, shadow, outline

    # Customize HUD class default values just to test HUD() without args
    HUD.FONT, HUD.SIZE, HUD.BOLD, HUD.ITALIC = ui.font, ui.size, ui.bold, ui.italic
    HUD.FGCOLOR, HUD.SHADOWCOLOR = ui.fgcolor, ui.shadow

    # Make our HUD
    hud = MyHUD()

    running = True
    ticks = 0

    while running:
        clock.tick(max_fps)
        ticks += 1

        for e in pygame.event.get():
            if e.type == MOUSEMOTION:
                mouse_pos = e.pos
            elif e.type == KEYDOWN:
                if e.key in (K_ESCAPE, K_q):
                    running = False
                elif not ui.is_interactive:
                    ui.set_interactive(hud)
                    pygame.display.set_caption('HUD Demo - See HUD for interactive keys | Q or ESC to quit')
                else:
                    ui.do_key(e.key, hud)
            elif e.type == QUIT:
                running = False

        # Update the spinner
        x = math.cos(angle) * radius + origin[0]
        y = math.sin(angle) * radius + origin[1]
        angle = (angle + 0.005) % (2 * math.pi)

        # Update the HUD values
        if ticks % 2 == 0:
            hud.update_all(
                ui.font, ui.size, ui.bold, ui.italic, ui.antialias,
                ui.fgcolor, ui.bgcolor, ui.gcolor,
                ui.colorkey, ui.shadow, ui.outline,
                clock.get_fps(), mouse_pos, math.degrees(angle))
        if ticks % max_fps == 0:
            hud.update_perf()

        # Draw
        screen.fill(clear_color)
        # background
        for p1, p2 in decorative_lines:
            pygame.draw.line(screen, Color('purple'), p1, p2)
        # spinner
        pygame.draw.line(screen, Color('white'), origin, (x, y))
        # hud
        hud.draw(screen)
        pygame.display.flip()


if __name__ == '__mian__':
    run_demo()
